/*
 * netopen.c
 *
 * Open a TCP network connection to the specified hostname and port.
 * This might need some porting work to work on Win32.
 */

#include "network.h"
#include <stdio.h>

#ifdef __WIN32__

#define _WIN32_WINNT 0x0501	/* Windows XP */
#include <windows.h>
#include <string.h>
#include <io.h>
#include <fcntl.h>
#include <winsock2.h>
#include <ws2tcpip.h>

static int winsock_started = 0;
int socket_errno;		/* For debuggability */

static void end_winsock(void)
{
  if ( winsock_started ) {
    WSACleanup();
    winsock_started = 0;
  }
}

static void start_winsock(void)
{
  WSADATA wd;

  if ( !winsock_started && !WSAStartup(MAKEWORD(2,0), &wd) ) {
    winsock_started = 1;
    atexit(end_winsock);
  }
}

static FILE *sock_to_file(SOCKET sock, int is_write)
{
  FILE *f;
  int fd;
  int osfhandle = sock;

  fd = _open_osfhandle(&osfhandle, is_write ? 0 : _O_RDONLY);
  if (fd < 0) {
    closesocket(fd);
    return NULL;
  }

  f = fdopen(fd, is_write ? "w" : "r");
  if (!f)
    _close(fd);

  return f;
}

#else
/* Unix, or something else which implements BSD sockets correctly */

#include <unistd.h>
#include <netdb.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>

/* Winsuck-compatibility */
typedef int SOCKET;
#define INVALID_SOCKET (-1)
#define start_winsock() do { } while (0)
#define closesocket(x) close(x)

static FILE *sock_to_file(SOCKET sock, int is_write)
{
  FILE *f;

  f = fdopen(sock, is_write ? "w" : "r");
  if (!f)
    close(sock);

  return f;
}

#endif

#ifdef AF_INET6

#ifndef AI_ADDRCONFIG
# define AI_ADDRCONFIG 0
#endif

static SOCKET open_socket(const char *name, int port)
{
  SOCKET sock;
  struct addrinfo *ai, *ap;
  struct addrinfo hints;
  char port_str[32];

  start_winsock();

  memset(&hints, 0, sizeof hints);
  hints.ai_family   = AF_UNSPEC;
  hints.ai_socktype = SOCK_STREAM;
  hints.ai_flags    = AI_ADDRCONFIG;

  sprintf(port_str, "%d", port);
  if (getaddrinfo(name, port_str, &hints, &ai))
    return INVALID_SOCKET;

  sock = INVALID_SOCKET;

  for (ap = ai; ap; ap = ap->ai_next) {
    sock = socket(ap->ai_family, ap->ai_socktype, ap->ai_protocol);
    if ( sock == INVALID_SOCKET )
      continue;

    if ( !connect(sock, ap->ai_addr, ap->ai_addrlen) )
      break;

    closesocket(sock);
    sock = INVALID_SOCKET;
  }

  freeaddrinfo(ai);

  return sock;
}

#else

/* Legacy IPv4-only API */

static SOCKET open_socket(const char *name, int port)
{
  struct hostent *he;
  SOCKET sock;
  struct in_addr **haddr;
  struct sockaddr_in si;

  start_winsock();

  he = gethostbyname(name);
  if ( !he || he->h_addrtype != AF_INET )
    return INVALID_SOCKET;

  sock = socket(PF_INET, SOCK_STREAM, 0);
  if ( sock == INVALID_SOCKET )
    return INVALID_SOCKET;

  for ( haddr = (struct in_addr **)he->h_addr_list ; *haddr ; haddr++ ) {
    memset(&si, 0, sizeof si);
    si.sin_family = AF_INET;
    si.sin_port = htons(port);
    memcpy(&si.sin_addr, *haddr, sizeof si.sin_addr);

    if ( !connect(sock, (struct sockaddr *)&si, sizeof si) )
      break;
  }
  if ( !*haddr ) {
    closesocket(sock);
    return INVALID_SOCKET;
  }

  return sock;
}

#endif

FILE *fopen_network(const char *name, int port, int is_write)
{
  SOCKET sock;

  sock = open_socket(name, port);
  if (sock == INVALID_SOCKET)
    return NULL;

  return sock_to_file(sock, is_write);
}
